from manim import *

from .polygon import BetterPolygon


class BetterRegularPolygon(BetterPolygon):
    def __init__(self, n: int, **kwargs):
        self.n = n
        regular_polygon = RegularPolygon(n=n, **kwargs).scale(2)
        super().__init__(*regular_polygon.get_vertices(), **kwargs)


from manim import *


class PiChart(VGroup):
    def __init__(self, data: dict[
        str, int
    ], colors: list, show_labels=True, **kwargs):
        super().__init__(**kwargs)
        self.data = data
        self.colors = colors
        self.show_labels = show_labels
        self.create_pi_chart()
    
    def create_pi_chart(self):
        total = sum([value for key, value in self.data.items()])
        angle = 0
        index = 0
        for key, value in self.data.items():
            color = self.colors[index] if self.colors else BLUE
            sector = Sector(
                arc_center=ORIGIN,
                start_angle=angle,
                angle=value/total*TAU,
                fill_color=color,
                fill_opacity=1,
                stroke_width=0
            )

            if self.show_labels:
                outward_direction = rotate_vector(RIGHT, angle + value/total*TAU/2)
                label = MathTex(f"{key} ({value})", color=color).scale(0.5)
                label.next_to(sector.get_center() + outward_direction*0.5, outward_direction)
                self.add(label)

            self.add(sector)
            angle += value/total*TAU
            index += 1


from manim import *


class ForceDiagram(VGroup):
    def __init__(self, object, **kwargs):
        self.forces = []
        self.object = object
        self.force_arrows = VGroup()
        super().__init__(self.force_arrows, **kwargs)

    def add_force(self, magnitude, angle, text = "", text_side = RIGHT, equal_length = False):
        group = VGroup()
        force = magnitude * np.array([np.cos(angle), np.sin(angle), 0])
        arrow = Arrow(
            start=self.object.get_center(),
            end=self.object.get_center() + force,
            buff=0,
        )

        if equal_length:
            midpoint = (arrow.get_start() + arrow.get_end()) / 2
            line = Line(
                start=midpoint + 0.1 * np.array([np.sin(angle), -np.cos(angle), 0]),
                end=midpoint - 0.1 * np.array([np.sin(angle), -np.cos(angle), 0]),
                color=WHITE
            )
            group.add(line)

        label = MathTex(text).next_to(arrow, text_side)
        self.forces.append({
            "vector": force,
            "label": label
        })
        self.force_arrows.add(arrow)

        group.add(arrow, label)
        self.add(group)
        return group

    def remove_force(self, index):
        self.forces.pop(index)
        self.force_arrows.remove(self.force_arrows[index])


from io import BytesIO

import numpy as np
import requests
from manim import ImageMobject
from PIL import Image


class WebImage(ImageMobject):
    def __init__(self, url, **kwargs):
        self.url = url
        arr = self.load_image()
        super().__init__(arr, **kwargs)

    def load_image(self):
        response = requests.get(self.url)
        image = Image.open(BytesIO(response.content))
        np_array = np.array(image)
        return np_array


from enum import Enum

import numpy as np
from manim import *


class CardSuit(Enum):
    HEARTS = 0
    DIAMONDS = 1
    CLUBS = 2
    SPADES = 3

class Card(Group):
    def __init__(self, suit: CardSuit, number: int = 1, face_up: bool = True, card_radius=0.1, **kwargs):
        super().__init__(**kwargs)
        self.suit = suit
        self.number = number
        self.face_up = face_up
        self.card_radius = card_radius
        
        card = RoundedRectangle(width=2, height=3, corner_radius=card_radius, color=WHITE, fill_opacity=1)
        
        suit_element1 = self.get_suit(30)
        suit_element1.move_to(LEFT * 0.7 + UP * 0.9 + card.get_center())

        number_element1 = self.get_number()
        number_element1.move_to(LEFT * 0.7 + UP * 1.25 + card.get_center())

        suit_element2 = self.get_suit(30)
        suit_element2.rotate(PI)
        suit_element2.move_to(RIGHT * 0.7 + DOWN * 0.9 + card.get_center())

        number_element2 = self.get_number()
        number_element2.rotate(PI)
        number_element2.move_to(RIGHT * 0.7 + DOWN * 1.25 + card.get_center())

        picture = self.get_picture()
        picture.move_to(card.get_center())

        self.back = self.get_back()
        self.back.set_opacity(0)

        self.add(card, suit_element1, number_element1, suit_element2, number_element2, picture, self.back)

    def get_suit(self, size=45):
        if self.suit == CardSuit.HEARTS:
            return Text("♥", font_size=size, color=RED)
        elif self.suit == CardSuit.DIAMONDS:
            return Text("♦", font_size=size, color=RED)
        elif self.suit == CardSuit.CLUBS:
            return Text("♣", font_size=size, color=BLACK)
        elif self.suit == CardSuit.SPADES:
            return Text("♠", font_size=size, color=BLACK)
    
    def get_number(self):
        return Text(str(self.format_number()), font_size=26, color=BLACK)
    
    def get_picture(self):
        group = Group()
        group.add(self.get_suit())

        return group
    
    def get_back(self):
        return RoundedRectangle(width=2, height=3, fill_color=BLUE, corner_radius=self.card_radius, fill_opacity=1)
    
    def format_number(self):
        if self.number == 1:
            return "A"
        elif self.number == 11:
            return "J"
        elif self.number == 12:
            return "Q"
        elif self.number == 13:
            return "K"
        else:
            return str(self.number)

def Flip(card: Card, **kwargs):
    card.face_up = not card.face_up
    def update(group, alpha):
        if (card.face_up):
            group.back.set_opacity(0)
        else:
            group.back.set_opacity(1)
    return AnimationGroup([Rotate(card, PI, axis=np.array([0, 1, 0]), **kwargs), UpdateFromAlphaFunc(card, update, **kwargs)], lag_ratio=0.5)



from manim import *


class BetterPolygon(Polygon):
    def __init__(self, *vertices, **kwargs):
        self.angles = []
        super().__init__(*vertices, **kwargs)

    def get_points(self, point_labels: list[str] = None, label_size: int = 24) -> VGroup:
        self.points = []

        for i, vertex in enumerate(self.get_vertices()):
            label = point_labels[i] if point_labels and i < len(point_labels) else None
            point = self.get_point(vertex, label, label_size)
            self.points.append(point)
        
        return VGroup(*self.points)

    def get_angles(self, angle_labels: list[str] = None, radius: float = 0.5, label_size: int = 24, label_as_angle=False) -> VGroup:
        self.angles = []

        vertices = self.get_vertices()
        num_vertices = len(vertices)
        for i in range(num_vertices):
            label = angle_labels[i] if angle_labels and i < len(angle_labels) else None
            angle = self.get_angle(i, label, radius, label_size, label_as_angle)
            self.angles.append(angle)
        
        return VGroup(*self.angles)
    
    def get_angle(self, index: int, label: str = None, radius: float = 0.5, label_size: int = 24, label_as_angle=False) -> VGroup:
        vertices = self.get_vertices()
        num_vertices = len(vertices)
        
        p1 = vertices[(index - 1) % num_vertices]
        p2 = vertices[index]
        p3 = vertices[(index + 1) % num_vertices]

        v1 = p1 - p2
        v2 = p3 - p2
        v1_unit = v1 / np.linalg.norm(v1)
        v2_unit = v2 / np.linalg.norm(v2)

        angle1 = np.arctan2(v1[1], v1[0])
        angle2 = np.arctan2(v2[1], v2[0])
        
        interior_angle = (angle2 - angle1) % (2 * PI)
        
        if interior_angle > PI:
            interior_angle = 2 * PI - interior_angle
            angle1, angle2 = angle2, angle1

        is_right_angle = abs(interior_angle - PI/2) < 0.01
        
        angle_marker = VGroup()
        
        if is_right_angle:
            right_angle_size = radius * 0.7
            
            # Calculate points for the right angle marker
            point1 = p2 + right_angle_size * v1_unit
            point2 = p2 + right_angle_size * v2_unit
            corner = p2 + right_angle_size * (v1_unit + v2_unit)
            
            # Create the L-shaped marker
            line1 = Line(point1, corner, color=WHITE)
            line2 = Line(corner, point2, color=WHITE)
            
            angle_marker.add(line1, line2)
        else:
            arc = Arc(
                radius=radius,
                start_angle=angle1,
                angle=interior_angle,
                arc_center=p2,
                color=WHITE
            )
            angle_marker.add(arc)
        
        if label or label_as_angle:
            mid_angle = angle1 + interior_angle / 2
            label_pos = p2 + 1.75 * radius * np.array([
                np.cos(mid_angle),
                np.sin(mid_angle),
                0
            ])

            content = f"{int(np.degrees(interior_angle))}" if label_as_angle else label
            text = MathTex(content, color=self.get_color(), font_size=label_size)
            text.move_to(label_pos)
            angle_marker.add(text)
        
        return angle_marker
    
    def get_side_labels(self, labels: list[str], label_size: int = 24) -> VGroup:
        vertices = self.get_vertices()
        num_vertices = len(vertices)
        
        side_labels = VGroup()
        for i in range(num_vertices):
            p1 = vertices[i]
            p2 = vertices[(i + 1) % num_vertices]
            
            mid_point = (p1 + p2) / 2
            outward_normal = np.array([-(p2[1] - p1[1]), p2[0] - p1[0], 0])
            outward_normal /= np.linalg.norm(outward_normal)
            mid_point -= 0.5 * outward_normal
            
            content = labels[i] if labels and i < len(labels) else f"l_{i}"
            text = MathTex(content, color=self.get_color(), font_size=label_size)
            text.move_to(mid_point)
            side_labels.add(text)
        
        return side_labels
    
    def get_start_angle(self, p1: np.ndarray, p2: np.ndarray) -> tuple[int, int]:
        v1 = p1 - p2
        return np.arctan2(v1[1], v1[0])


from manim import *


class ParticleSimulation(Group):
    def __init__(self, container_dimensions = (1, 1), n_particles=100, temp=273, particle_colors=[BLUE,WHITE,RED], **kwargs):
        super().__init__(**kwargs)
        self.container_dimensions = container_dimensions
        self.container = Rectangle(width=container_dimensions[0] + 0.2, height=container_dimensions[1] + 0.2, stroke_width=1)
        self.container.shift(RIGHT * 0.5 + UP * 0.5)
        self.add(self.container)

        self.particles = []
        self.temp = temp
        for i in range(n_particles):
            col = particle_colors[i % len(particle_colors)]
            particle = Dot(radius=0.05, color=col)
            pos = np.array([np.random.rand() * container_dimensions[0] * 0.5, np.random.rand() * container_dimensions[1] * 0.5, 0])
            particle.move_to(np.array([pos[0] + (self.container.get_center()[0] - (self.container_dimensions[0] / 4)), pos[1] + (self.container.get_center()[1] - (self.container_dimensions[1] / 4)), 0]))
            self.particles.append({
                "particle": particle,
                "pos": pos,
                "velocity": [np.random.rand() for _ in range(2)]
            })
            self.add(particle)

    def update_particles(self, dt = 1 / config.frame_rate):
        for particle in self.particles:
            pos = particle["pos"]
            velocity = particle["velocity"]
            for i in range(2):
                pos[i] += velocity[i] * dt * (self.temp / 273 /2)
                if pos[i] <= 0 or pos[i] >= self.container_dimensions[i]:
                    velocity[i] *= -1
            particle["pos"] = pos
            particle["velocity"] = velocity
            particle["particle"].move_to(np.array([pos[0] + (self.container.get_center()[0] - (self.container_dimensions[0] / 2)), pos[1] + (self.container.get_center()[1] - (self.container_dimensions[1] / 2)), 0]))

    def set_temp(self, temp):
        self.temp = temp


from manim import *


class SlopeField(VGroup):
    def __init__(self, func, x_range=(-10, 10), y_range=(-10, 10), step_size=1, **kwargs):
        self.func = func
        self.x_range = x_range
        self.y_range = y_range
        self.step_size = step_size
        self.axes = self.get_axes()
        self.field = self.get_field()
        super().__init__(self.axes, self.field, **kwargs)

    def get_axes(self):
        axes = Axes(
            x_range=self.x_range,
            y_range=self.y_range,
            axis_config={"include_tip": False},
        )
        axes.center()
        return axes

    def get_field(self):
        f = lambda p: np.array([1, self.get_slope(p[0], 0.000001), 0])
        field = ArrowVectorField(f)
        
        field.set_opacity(0.5)
        field.set_color(WHITE)
        field.length_func = lambda norm: 0.1

        field.move_to(self.axes)
        field.fit_to_coordinate_system(self.axes)
        return field
    
    def get_slope(self, x, dx):
        return (self.func(x + dx) - self.func(x)) / dx


from manim import *


class RotatingFunction(VMobject):
    def __init__(self, func, axis=[1, 0], color=BLUE, opacity=0.5, x_range=[-1, 1], **kwargs):
        super().__init__(**kwargs)
        self.func = func

        self.axes = ThreeDAxes()
        self.add(self.axes)

        axis = np.array(axis, dtype=float)
        if np.allclose(axis, [0, 1]):
            axis = -axis
        self.axis = axis / np.linalg.norm(axis)

        self.graph = self.axes.plot(func, x_range=x_range)
        self.add(self.graph)

        self.angle_tracker = ValueTracker(0)
        
        # based on the axis passed in
        if np.allclose(self.axis, [1, 0]) or np.allclose(self.axis, [-1, 0]):
            self.surface_function = lambda u, v: np.array([
                u,
                self.func(u) * np.cos(v),
                self.func(u) * np.sin(v),
            ])
        elif np.allclose(self.axis, [0, 1]) or np.allclose(self.axis, [0, -1]):
            self.surface_function = lambda u, v: np.array([
                u * np.cos(v),
                self.func(u),
                u * np.sin(v),
            ])
        else:
            raise ValueError("Axis must be one of [1, 0], [0, 1]")

        self.surface = always_redraw(lambda: Surface(
            self.surface_function,
            u_range=x_range,
            v_range=[0, self.angle_tracker.get_value()],
            resolution=(16, 32),
            fill_opacity=opacity,
            fill_color=color,
            stroke_width=0,
            checkerboard_colors=False
        ))
        self.add(self.surface)

    def rotate(self, angle=2*PI):
        return AnimationGroup(
            self.angle_tracker.animate.set_value(angle),
            Rotate(
                self.graph,
                angle=angle,
                axis=np.array([self.axis[0], self.axis[1], 0]),
                about_point=ORIGIN
            )
        )


from .card import *
from .force_diagram import *
from .geometry import *
from .particle_simulation import *
from .pi_chart import *
from .rotating_function import *
from .slope_field import *
from .web_image import *
